/* $Id: check_ps.c,v 1.35 1998/06/26 19:32:45 dps Exp $ */
/* The main detection code, */
/* Just runs this the background and detects nasty version of ps */
/* (c) D.P.Simpson 1997, rediristibution and use encouraged, provided */
/* you comply with the GPL. */

#include "config.h"
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif /* HAVE_CTYPE_H */
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif /* HAVE_SIGNAL_H */
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif /* HAVE_ERRNO_H */
#ifdef HAVE_SETJMP_H
#include <setjmp.h>
#endif /* HAVE_SETJMP_H */
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif /* HAVE_SYS_STAT_H */
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif /* HAVE_SYS_WAIT_H */
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif /* HAVE_SYS_TIME_H */
#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif /* HAVE_SYS_RESOURCE_H */
#include "prog.h"
#include "cfg_check.h"
#include "numbuf.h"

#ifndef HAVE_LSTAT
#define lstat(x,y) stat(x,y)
#endif /* HAVE_LSTAT */

/* Child catcher */
static RETSIGTYPE catch_child(int sig)
{
    int status;
    pid_t pid;

    sig=sig;
    pid=wait(&status);
    if (WIFSTOPPED(status))
       kill(pid, SIGCONT);     /* Stop crackers stoping the ps process */
                               /* (major paranoia) */
    return;
}

/* Test a process id */
static int process_ok(numbuf *pids, int pid, const numbuf *fpl, numbuf *nl,
		      int (*current_proc)(int))
{
    int i;

    for (i=0; i<pids->pos; i++)
    {
	if (pids->nums[i]==pid)
	{
	    pids->nums[i]=-2;	/* Used, OK */
	    return 1;		/* I am happy */
	}
    }

    if (!(*current_proc)(pid))
    {
	/* Check the fake pid list */
	for (i=0; i<fpl->pos; i++)
	{
	    if (fpl->nums[i]==pid)
	    {
#ifdef DEBUG
		fprintf(stderr, "Hit...fake pid %d\n", pid);
#endif		
		log("Fake pid %d detected", pid);
	    }
	}
	/* Add to list of fake pids */
	if (nl->pos==nl->max && expand_buf(nl)==0)
	{
	    log("Fake pid list overflow, size %d", nl->pos);
	}
	nl->nums[(nl->pos)++]=pid;
	return 2;		/* Already dead. ps probably tickles this. */
    }
    return 1;
}


static jmp_buf runps_env;
static RETSIGTYPE (*oldalrm)(int); 

static RETSIGTYPE runps_alarm(int sig)
{
    sig=sig;
    longjmp(runps_env,1);
    /* Not reached */
}

/* ps output analyser, mainly just toilet roll. Some anti-DoS stuff too
 * lest jokers write a DoS against this program version of ps. However
 * that would be obviously hacked so they might not take the risk. 
 */
static int run_ps(numbuf *pids, const numbuf *fpl, numbuf *nl,
		  int (*current_proc)(int))
{
    int max, pos;
    int fd[2], null;
    FILE *in;
    int c, num, nc, i;
    volatile pid_t pid;
    gid_t gid;
    static const struct rlimit pid_lim={RLIMIT_NPROC, MAX_PROCS};
    static const struct rlimit mlock_lim={RLIMIT_MEMLOCK, 0};
    enum {SKIP_WS, GET_NUM, SKIP_END, DOS_ATTACK} line;
    int retest=0;
#ifdef PARANIOD
    struct stat stbuf, stbuf2, stbuf3;
    int do_chmod;
    FILE *f;
#endif /* PARNIOD */

    signal(SIGCHLD, catch_child);

    if ((null=open("/dev/null", O_RDWR))==-1)
    {
       log("Could not open /dev/null (%s)", strerror(errno));
       return 0;
    }

    if (pipe(fd))
    {
       log("Could not create pipe");
       close(null);
       return 0;
    }

    if ((in=fdopen(fd[0],"r"))==NULL)
    {
       log("fdopen failure");
       close(null);
       close(fd[0]);
       close(fd[1]);
       return 0;
    }

    /* Use alarm to stop ps==crack working nicely for crackers */
    pid=0;
    oldalrm=signal(SIGALRM, SIG_IGN);
    if (setjmp(runps_env)!=0)
    {
	if (pid!=0)
	    kill(pid, SIGKILL);	/* Kill ps */
	alarm(0);		/* Disable alarm */
	signal(SIGALRM, oldalrm); /* Reset signal handling */
	log("ps output timeout"); /* Nasty ps */
	log_flush();		/* Tell admin */
	sleep(DISABLE_TIME);	/* Sleep awhile, lest ps is password cracker */
	return 0;		/* Failed */
    }
    signal(SIGALRM, runps_alarm);
    alarm(PS_MAXTIME);

    switch(pid=fork())
    {
    case -1:
       log("Fork failed");
       close(null);
       close(fd[0]);
       close(fd[1]);
       return 0;

    case 0: /* child */
       signal(SIGTSTP, SIG_IGN); /* Stop TSTP */

       /* I know this is not race condition proof but worth soemthing
	* anyway (should make life harder for crackers...) */
#ifdef PARANIOD
       do_chmod=0;
       if (lstat(quote(PS_PATH), &stbuf))
       {
	   log("Stat " quote(PS_PATH) " failed\n");
	   exit(0);
       }
       if ((f=fopen(quote(PS_PATH), "r"))==NULL)
       {
	   log("Could not open " quote (PS_PATH));
	   exit(0);
       }
       if (fstat(fileno(f), &stbuf2))
       {
	   fclose(f);
	   log("Stat " quote(PS_PATH) " as openned file failed");
	   exit(0);
       }
       if (!S_ISREG(stbuf.st_mode) || !S_ISREG(stbuf2.st_mode))
       {
	   log(quote(PS_PATH) " is not a regular file");
	   exit(0);
       }

       /* Ultra-paranoia check */
       if (stbuf.st_dev!=stbuf2.st_dev || stbuf.st_ino!=stbuf2.st_ino)
       {
	   log (quote(PS_PATH) " had dev/ino=%d/%d file had %d/%d",
		stbuf.st_dev, stbuf.st_ino, stbuf2.st_dev, stbuf2.st_ino);
	   exit(0);
       }
#ifdef KILLSUID
       if (stbuf2.st_mode & S_ISUID)
       {
	   log("Killing SUID bit of " quote(PS_PATH));
	   do_chmod=1;
	   stbuf2.st_mode &= ~S_ISUID;
       }
#endif /* KILLSUID */
#ifdef KILLSGID
       if (stbuf2.st_mode & S_ISGID)
       {
	   log("Killing SGID bit of " quote(PS_PATH));
	   do_chmod=1;
	   stbuf2.st_mode &= ~S_ISGID;
       }
#endif /* KILLSGID */
#ifdef HAVE_FCHMOD
       if (do_chmod)
	   fchmod(fileno(f), stbuf2.st_mode);
#endif /* HAVE_FCHMOD */
       fclose(f);
#ifndef HAVE_FCHMOD
       /* Race-able but better than nothing */
       if (do_chmod)
	   chmod(quote(PS_PATH), stbuf2.st_mode);
#endif /* not HAVE_FCHMOD */

       /* Final paranioa check... */
       if (stat(quote(PS_PATH), &stbuf3))
       {
	   log(quote(PS_PATH) " stat failed on third try");
	   exit(0);
       }
       if (stbuf3.st_mode!=stbuf2.st_mode)
       {
	   log("Permission changed since I opened " quote(PS_PATH));
	   exit(0);
       }

       /* We know that stbuf2.st_dev==stbuf.st_dev and
	* stbuf2.st_ino==stbuf.st_ino here. */
       if (stbuf3.st_dev!=stbuf.st_dev || stbuf3.st_ino!=stbuf.st_ino)
       {
	   log (quote(PS_PATH) " changed from dev/ino=%d/%d to %d/%d",
		stbuf.st_dev, stbuf.st_ino, stbuf3.st_dev, stbuf3.st_ino);
	   exit(0);
       }
#endif /* PARANIOD */

       close(fd[0]);
       close(0);
       close(1);
       dup2(null,0);
       dup2(fd[1],1);
#ifndef DEBUG
       close(2);
       dup2(null,2);
#endif
       

#ifdef SETRLIMIT
       /* set limits */
       setrlimit(RLIMIT_NPROC, (struct rlimit *) &pid_lim);
       setrlimit(RLIMIT_MEMLOCK, (struct rlimit *) &mlock_lim);
#endif

       /* Drop privs */
       gid=DEFGID;
       setregid(DEFGID, DEFGID); /* Set gid */
#ifdef SETGROUPS
       setgroups(1, &gid);	 /* Kill extra groups */
#endif
       setreuid(DEFUID, DEFUID); /* Set uid */

       /* Run ps */
       execl(quote(PS_PATH),"ps",quote(PS_ARGS),NULL);
       exit(0);			 /* Should never get here */

    default:
       close(fd[1]);
       close(null);
       break;
    }

    line=SKIP_END;		/* Skip 1st line */
    nc=0;			/* character count is zero */
    while(line!=DOS_ATTACK && (c=fgetc(in))!=EOF)
    {
       switch(line)
       {
       case SKIP_WS:
           if (isspace(c))
	   {
	       nc++;
	       if (nc>TOLERANCE) /* Test for DoS attack... */
	       {
		   log("Attempted DoS attack: >" quote(TOLERANCE)
		       " bytes leading white space");
		   line=DOS_ATTACK;
		   break;
	       }
               continue;
	   }
           num=0;
	   nc=0;
           line=GET_NUM;
           /* FALL THRU */

       case GET_NUM:
           if (isdigit(c))
           {
	       nc++;
	       if (nc>NUM_TOLERANCE)
	       {
		   log("Attempted  DoS attack: >" quote(NUM_TOLERANCE)
		       " digits in pid");
		   line=DOS_ATTACK;
		   break;
	       }
               num=num*10+(c-'0');
               continue;
           }
           if (nc==0)
	   {
	       log("Attempt to attack me by not starting with a number");
	   }
	   else if (num!=pid)
	   {
	       switch(nc=process_ok(pids, num, fpl, nl, current_proc))
	       {
	       case 0:
		   nuke_proc(num); /* Ask no questions, nuke */
		   retest=1;
		   break;

	       case 1:
		   break;

	       case 2:
#ifdef DEBUG
		   fprintf(stderr, "retest %d\n", num);
#endif
		   retest=1;
		   break;

	       default:
		   log("process ok returned %d", nc);
		   break;
	       }
	   }
	   nc=0;
	   line=SKIP_END;
	       /* FALL THRU */

       case SKIP_END:
	   nc++;
	   if (nc>TOLERANCE) /* Test for DoS attack... */
	   {
	       log("Attempted DoS attack: more than " quote(TOLERANCE)
		   " before newline");
	       line=DOS_ATTACK;
	       break;
	   }
           if (c=='\n')
               line=SKIP_WS;
           break;
	       
       case DOS_ATTACK:
	   break;

       default:
           log("Imposible state %d", line);
           fclose(in);
           return 0;
       }
    }
    kill(pid, SIGKILL);		/* Kill ps (stop ps DoS attack) */
    alarm(0);			/* Disable alarm */
    signal(SIGALRM, oldalrm);	/* Restore signal handler */
    /* No need to wait, SIGCHLD handler does that */

    /* Did ps tell the truth or was there something it was not telling us?
     * current_process is a fuinction to eliminate false positives due to
     * races (a process get found by get_pids but exits before we execute
     * ps). A small risk of a false -ve is introduced... */
    for (i=0; i<pids->pos; i++)
    {
	if (pids->nums[i]!=-2 && (*current_proc)(pids->nums[i]))
	{
	    if (nuke_proc(pids->nums[i]))
		    log("Hidden pid %d", pids->nums[i]);
	}
    }
    log_flush();		/* Flush log lines. */
    return (retest==0) ? 1 : 2;
}



/* demon body */
/* The demon uses *two* fake pid lists to avoid generating alarms due to race
 * conditions. If a fake pid turns up twice then it is not a race but somethig
 * rather nastier (thanks to bugtraq for the race and NASA for reminding me
 * that fake versioons of ps might add some things that should be around but
 * are not actually present). */
void demon_body(numbuf *(*pid_list)(void),  int (*current_proc)(int))
{
    int idx,r,did_check;
    numbuf *alrm[2], *nums;

    ignore_signals();
    alrm[0]=allocate();
    alrm[1]=allocate();
    idx=0;
    did_check=0;		/* Avoid spamming admin */
    while(1)
    {
       log_start();		/* Start recording period */
       if ((nums=(*pid_list)())==NULL)
       {
           log("pid_list failed");
	   log_flush();		/* Send logged lines to log */
           sleep (SHORTNAP);	/* Retry after a pause */
           continue;
       }
       r=run_ps(nums, alrm[idx], alrm[1-idx], current_proc); /* Runs ps */
       if (r==0)
       {
	   sleep(SHORTNAP);
	   continue;		/* Retry after a pause */
       }
       empty_nums(alrm[idx]);	/* Delete fake pid list */
       idx=1-idx;		/* Switch to bew fake pid list */
       free_nums(nums);		/* Free memory */
       log_flush();		/* Send logged liens to log */
       if (r==1 && did_check==0)
       {
	   did_check=1;		/* Set flag */
	   sleep(RECHECKNAP);	/* Sleep a bit */
	   continue;		/* Check ps again */
       }
       did_check=0;		/* Did check */
       sleep(LONGNAP);		/* Sleep alot */
    }
}

/* One shot verson of demon body */
int one_shot(numbuf *(*pid_list)(void), int (*current_proc)(int))
{
    int r, idx;
    numbuf *alrm[2], *nums;

    ignore_signals();
    log("Entered one_shot");
    alrm[0]=allocate();
    alrm[1]=allocate();
    idx=0;
    while(1)
    {
	log_start();		/* Start recording period */
	if ((nums=(*pid_list)())==NULL)
	{
	    log("pid_list failed");
	    log_flush();		/* Send logged lines to log */
	    sleep (SHORTNAP);	/* Retry after a pause */
	    continue;
}
	r=run_ps(nums, alrm[idx], alrm[1-idx], current_proc); /* Runs ps */
#ifdef DEBUG
	fprintf(stderr, "exit with r=%d\n", r);
#endif
	if (r==0)
	{
	    log("runps failed");
	    sleep(SHORTNAP);
	    continue;		/* Retry after a pause */
	}
	free_nums(nums);	/* Free memory */
	log_flush();		/* Send logged liens to log */
	if (r==1 || idx>0)
	    return (idx==0) ? 0 : 1;
	if (r!=1)
	{
	    log("Some pids seems to be fake");
	    idx=1-idx;
	    sleep(RECHECKNAP);
	    log("rechecking ps");
	    continue;
	}
    }
    fputs("Oops! Got to IP2\n", stderr);
    return 2;
}
